/****************************************************************
Copyright (c) 1998 Sandro Sigala <ssigala@globalnet.it>.
All rights reserved.

Permission to use, copy, modify, and distribute this software
and its documentation for any purpose and without fee is hereby
granted, provided that the above copyright notice appear in all
copies and that both that the copyright notice and this
permission notice and warranty disclaimer appear in supporting
documentation, and that the name of the author not be used in
advertising or publicity pertaining to distribution of the
software without specific, written prior permission.

The author disclaim all warranties with regard to this
software, including all implied warranties of merchantability
and fitness.  In no event shall the author be liable for any
special, indirect or consequential damages or any damages
whatsoever resulting from loss of use, data or profits, whether
in an action of contract, negligence or other tortious action,
arising out of or in connection with the use or performance of
this software.
****************************************************************/

#include "config.h"

#include <kapp.h>
#include <kmenubar.h>
#include <kmsgbox.h>
#include <kstatusbar.h>
#include <ktopwidget.h>
#include <qaccel.h>
#include <qfile.h>
#include <qpopmenu.h>
#include <qpushbt.h>
#include <qtstream.h>

#include "krepton.h"
#include "gamewindow.h"
#include "gamewidget.h"
#include "pwddialog.h"
#include "scoredialog.h"
#include "newscoredialog.h"
#include "splashwindow.h"

const int minimum_width = 5 * sprite_width;
const int minimum_height = 3 * sprite_height;
const int default_width = 9 * sprite_width;
const int default_height = 7 * sprite_height;

GameWindow::GameWindow(QWidget *, const char *name)
        : KTopLevelWidget(name)
{
	setCaption("Repton");

	newgame_popup = new QPopupMenu();
	continuegame_popup = new QPopupMenu();
	file_popup = new QPopupMenu();
	file_popup->insertItem("&New Game", newgame_popup);
	file_popup->insertItem("&Continue Game", continuegame_popup);
        pauseID = file_popup->insertItem("&Pause", this, SLOT(menu_pause()), Key_F2);
        file_popup->insertItem("&End Game", this, SLOT(menu_endGame()), CTRL+Key_E);
	file_popup->insertSeparator();
        file_popup->insertItem("&High Scores...", this, SLOT(menu_highScores()));
	file_popup->insertSeparator();
	file_popup->insertItem("&Quit", qApp, SLOT(quit()), CTRL+Key_Q);

	loadsprites_popup = new QPopupMenu();

	options_popup = new QPopupMenu();
	options_popup->insertItem("&Load Sprites", loadsprites_popup);
	options_popup->insertSeparator();
	soundsID = options_popup->insertItem("Sounds", this, SLOT(menu_sounds()));

	help_popup = new QPopupMenu();
	help_popup->insertItem("&Contents", this, SLOT(menu_help()), Key_F1);
	help_popup->insertSeparator();
	help_popup->insertItem("&About KRepton...", this, SLOT(menu_about()));
//	help_popup->insertItem("About &KDE...", this, SLOT(menu_aboutKDE()));

	loadEpisodeList(episodes, true);	// Global episodes
	loadEpisodeList(episodes, false);	// Local episodes

	Episode *e; int i;
	bool global = true;
	for (e = episodes.first(), i = 0; e != 0; e = episodes.next(), ++i) {
		if (!e->global && global) {
			global = false;
			newgame_popup->insertSeparator();
			continuegame_popup->insertSeparator();
			loadsprites_popup->insertSeparator();
		}
		e->new_game_id = newgame_popup->insertItem(e->name, i + 100);
		e->continue_game_id = continuegame_popup->insertItem(e->name, i + 200);
		e->load_sprites_id = loadsprites_popup->insertItem(e->name, i + 300);
	}

	menu = new KMenuBar(this);
	menu->insertItem("&File", file_popup);
	menu->insertItem("&Options", options_popup);
	menu->insertItem("&Help", help_popup);
	setMenu(menu);
	connect(menu, SIGNAL(moved(menuPosition)),
		this, SLOT(movedMenu(menuPosition)));
	connect(menu, SIGNAL(activated(int)),
		this, SLOT(activatedMenu(int)));

	status = new KStatusBar(this);
	status->insertItem("Diamonds: 999", 1);
	status->insertItem("Time: 9999", 2);
	status->insertItem("Score: 999999", 3);
	status->insertItem("Key", 4);	// A icon in the future
	setStatusBar(status);
	status->changeItem("", 1);
	status->changeItem("", 2);
	status->changeItem("", 3);
	status->changeItem("", 4);

	game = new GameWidget(this);
	setView(game);
	connect(game, SIGNAL(changedStats(int, int, int, bool)),
		this, SLOT(updateStats(int, int, int, bool)));
	connect(game, SIGNAL(gameOver()), this, SLOT(gameOver()));

	setMinimumSize(minimum_width, minimum_height + menu->height() + status->height());
	setSizeIncrement(sprite_width, sprite_height);
	resize(default_width, default_height + menu->height() + status->height());

	KConfig *config = kapp->getConfig();
	config->setGroup("Options");
	game->do_sounds = config->readBoolEntry("Sounds", true);
	options_popup->setItemChecked(soundsID, game->do_sounds);

	show();
	SplashWindow *splash = new SplashWindow();
	splash->show();
}

void GameWindow::menu_pause()
{
	if (game->in_game) {
		game->in_pause = !game->in_pause;
		file_popup->setItemChecked(pauseID, game->in_pause);
		game->repaint(false);
	}
}

void GameWindow::menu_endGame()
{
	if (game->in_game) {
		game->in_game = false;
		game->repaint(false);
		file_popup->setItemChecked(pauseID, false);
		gameOver();
	}
}

void GameWindow::menu_highScores()
{
	ScoreDialog d(episodes, this);
	d.exec();
}

void GameWindow::menu_sounds()
{
	game->do_sounds = !game->do_sounds;
	options_popup->setItemChecked(soundsID, game->do_sounds);
	KConfig *config = kapp->getConfig();
	config->setGroup("Options");
	config->writeEntry("Sounds", game->do_sounds);
}

void GameWindow::menu_help()
{
	kapp->invokeHTMLHelp("", "");
}

void GameWindow::menu_about()
{
	KMsgBox::message(this, "About KRepton",
			 "KRepton " VERSION "\n\n"
			 "by Sandro Sigala <ssigala@globalnet.it>\n\n"
			 "Feedbacks are welcome!\n",
			 KMsgBox::INFORMATION);
}

void GameWindow::menu_aboutKDE()
{
}

void GameWindow::updateStats(int diamonds, int secs, int points, bool key)
{
	QString d, s, p;
	d.setNum(diamonds);
	s.setNum(secs);
	p.setNum(points);
	status->changeItem(QString("Diamonds: ") + d, 1);
	status->changeItem(QString("Time: ") + s, 2);
	status->changeItem(QString("Score: ") + p, 3);
	// The key status may be replaced by a icon in the future
	status->changeItem(key ? "Key" : "", 4);
}

void GameWindow::gameOver()
{
	status->changeItem("", 1);
	status->changeItem("", 2);
	status->changeItem("", 3);
	status->changeItem("", 4);

	KConfig *config = kapp->getConfig();
	config->setGroup("High Score");
	QString s, score;
	score = config->readEntry(current_episode->filename + "Score", "0");
	if (game->current_map->num_points > score.toInt()) {
		NewScoreDialog d(this);
		if (d.exec()) {
			score.setNum(game->current_map->num_points);
			config->writeEntry(current_episode->filename + "Score", score);
			config->writeEntry(current_episode->filename + "Name", d.name());

			ScoreDialog h(episodes, this);
			h.exec();
		}
	}
}

void GameWindow::movedMenu(menuPosition pos)
{
	if (pos == Top || pos == Bottom) {
		setMinimumSize(minimum_width, minimum_height + menu->height() + status->height());
		resize(width(), height() + menu->height());
	} else {
		setMinimumSize(minimum_width, minimum_height + status->height());
		resize(width(), height() - menu->height());
	}
}

void GameWindow::activatedMenu(int id)
{
	for (Episode *e = episodes.first(); e != 0; e = episodes.next()) {
		QString path;
		if (e->global)
			path = kapp->kde_datadir() + "/krepton/data/" + e->filename;
		else
			path = kapp->localkdedir() + "/share/apps/krepton/data/" + e->filename;

		if (e->new_game_id == id) {
			current_episode = e;
			game->in_game = false;
			file_popup->setItemChecked(pauseID, false);
			game->loadEpisode(path);
			game->startGame(e->name, 0);
			break;
		} else if (e->continue_game_id == id) {
			current_episode = e;
			game->in_game = false;
			file_popup->setItemChecked(pauseID, false);
			game->loadEpisode(path);
			PwdDialog d(this);
			if (d.exec()) {
				QString pwd = d.password();
				pwd = pwd.stripWhiteSpace();
				if (pwd == "")
					KMsgBox::message(this, "KRepton", "No idea?");
				else
					game->startGamePassword(e->name, pwd);
			}
			break;
		} else if (e->load_sprites_id == id) {
			if (game->in_game) {
				game->loadSprites(path);
				game->repaint(false);
			}
			break;
		}
	}
}
