/****************************************************************
Copyright (c) 1998 Sandro Sigala <ssigala@globalnet.it>.
All rights reserved.

Permission to use, copy, modify, and distribute this software
and its documentation for any purpose and without fee is hereby
granted, provided that the above copyright notice appear in all
copies and that both that the copyright notice and this
permission notice and warranty disclaimer appear in supporting
documentation, and that the name of the author not be used in
advertising or publicity pertaining to distribution of the
software without specific, written prior permission.

The author disclaim all warranties with regard to this
software, including all implied warranties of merchantability
and fitness.  In no event shall the author be liable for any
special, indirect or consequential damages or any damages
whatsoever resulting from loss of use, data or profits, whether
in an action of contract, negligence or other tortious action,
arising out of or in connection with the use or performance of
this software.
****************************************************************/

#include "config.h"

#include <mediatool.h>
#include <kaudio.h>
#include <kapp.h>
#include <qfile.h>
#include <qtstream.h>
#include <stdlib.h>
#include <math.h>

#include "krepton.h"
#include "map.h"

Map::Map()
{
	monsters.setAutoDelete(true);
	transporters.setAutoDelete(true);
	status_die = false;
	status_end_level = false;
	audio = new KAudio();
}

Map::~Map()
{
	deleteTransporters();
	deleteMonsters();
	delete [] data;
	delete audio;
}

void Map::loadData(const QString& filename)
{
	QFile f(filename);
	if (!f.open(IO_ReadOnly))
		fatal("Cannot open data files.\nHave you correctly installed KRepton?");

	QTextStream t(&f);
	int num_transporters;
	t >> width >> height >> num_transporters >> num_secs;
	t >> password;

	while (num_transporters--) {
		int ox, oy, dx, dy;
		t >> ox >> oy >> dx >> dy;
		addTransporter(ox, oy, dx, dy);
	}

	data = new SpriteObj[width * height];

	num_diamonds = num_points = 0;
	have_key = false;
	sprite = Obj_Repton;

	bool got_position = false;
	for (int y = 0; y < height; ++y) {
		char c;
		for (int x = 0; x < width; ++x) {
			t >> c;
			ref(x, y) = (SpriteObj)(c - 'A');
			switch (ref(x, y)) {
			case Obj_Blip:
				ref(x, y) = Obj_Empty;
				addMonster(x, y, Obj_Blip);
				break;
			case Obj_Strongbox:
			case Obj_Cage:
			case Obj_Diamond:
				++num_diamonds;
				break;
			case Obj_Plant:
				addMonster(x, y, Obj_Plant);
				break;
			case Obj_Repton:
				if (got_position)
					fatal("Duplicate Repton position");
				got_position = true;
				xpos = x;
				ypos = y;
				break;
			default:;	// Avoid warning
			}
		}
	}
	if (!got_position)
		fatal("No Repton position");

	f.close();

	prepareMonsters();
}

void Map::playSound(Sound s)
{
	if (!*do_sounds || audio->serverStatus() != 0)
		return;

	QString name;
	switch (s) {
	case Sound_Die:
		name = "die.wav";
		break;
	case Sound_Got_Diamond:
		name = "diamond.wav";
		break;
	case Sound_Got_Crown:
		name = "crown.wav";
		break;
	case Sound_Cage_Blip:
		name = "cage.wav";
		break;
	case Sound_Got_Key:
		name = "key.wav";
		break;
	case Sound_Got_Time:
		name = "time.wav";
		break;
	case Sound_Transport:
		name = "transport.wav";
		break;
	case Sound_Kill_Monster:
		name = "monster.wav";
		break;
	case Sound_Broken_Egg:
		name = "egg.wav";
		break;
	}

	QString fname = kapp->kde_datadir() + "/krepton/wav/" + name;
//	audio->stop();
	audio->play(fname);
//	audio->sync();
}

void Map::addTransporter(int ox, int oy, int dx, int dy)
{
	transporters.append(new Transporter(ox, oy, dx, dy));
}

Transporter *Map::findTransporter(int x, int y)
{
	Transporter *t;
	for (t = transporters.first(); t != 0; t = transporters.next())
		if (t->orig_x == x && t->orig_y == y)
			return t;
	fatal("Invalid transporter");
	return 0;
}

void Map::deleteTransporters()
{
	Transporter *t;
	while ((t = transporters.first()) != 0)
		transporters.remove();
}

void Map::addMonster(int x, int y, SpriteObj type)
{
	monsters.append(new Monster(x, y, type));
}

void Map::prepareMonsters()
{
	Monster *m;
	for (m = monsters.first(); m != 0; m = monsters.next()) {
		int x = m->xpos, y = m->ypos;
		if (m->type == Obj_Blip) {
			if (!isempty(x-1, y) && isempty(x, y-1))
				m->orientation = North;
			else if (!isempty(x, y-1) && isempty(x+1, y))
				m->orientation = East;	
			else if (!isempty(x+1, y) && isempty(x, y+1))
				m->orientation = South;
			else if (!isempty(x, y+1) && isempty(x-1, y))
				m->orientation = West;
		}
	}
}

Monster *Map::findMonster(int x, int y)
{
	Monster *m;
	for (m = monsters.first(); m != 0; m = monsters.next())
		if (m->xpos == x && m->ypos == y)
			return m;
	return 0;
}

void Map::deleteMonsters()
{
	Monster *m;
	while ((m = monsters.first()) != 0)
		monsters.remove();
}

// Try falling down.
bool Map::tryFallDown(int x, int y)
{
	SpriteObj obj = xy(x, y);

	if (xy(x, y+1) != Obj_Empty)
		return false;

	ref(x, y) = Obj_Empty;
	Monster *m;
	if ((m = findMonster(x, y+1)) != 0 && m->type == Obj_Monster) {
		playSound(Sound_Kill_Monster);
		killMonster(m);
	}
	if (obj == Obj_Egg)
		ref(x, y+1) = Obj_Falling_Egg;
	else
		ref(x, y+1) = obj;
	if (xy(x, y+2) == Obj_Wall && (obj == Obj_Egg || obj == Obj_Falling_Egg)) {
		playSound(Sound_Broken_Egg);
		ref(x, y+1) = Obj_Broken_Egg;
	}
	if (xy(x, y+2) == Obj_Repton) {
		playSound(Sound_Die);
		status_die = true;
	}

	return true;
}

// Check if the specified object can fall to the left or to the right.
bool Map::tryFallHorizontal(int x, int y, int xd)
{
	SpriteObj obj = xy(x, y);

	if (xy(x+xd, y+1) == Obj_Empty && xy(x+xd, y) == Obj_Empty) {
		ref(x, y) = Obj_Empty;
		if (obj == Obj_Egg)
			ref(x+xd, y) = Obj_Falling_Egg;
		else
			ref(x+xd, y) = obj;
		return true;
	}

	return false;
}

void Map::tryFallLeftOrRight(int x, int y)
{
	SpriteObj obj = xy(x, y);

	switch(xy(x, y+1)) {
	case Obj_Rock:
	case Obj_Skull:
	case Obj_Diamond:
	case Obj_Egg:
	case Obj_Key:
	case Obj_Bomb:
	case Obj_Broken_Egg:
	case Obj_Broken_Egg1:
	case Obj_Broken_Egg2:
	case Obj_Broken_Egg3:
		// Curved objects.  Rocks and eggs can fall onto these objects.
		// If can't fall to the left, fall to the right.
		if (!tryFallHorizontal(x, y, -1))
			if (!tryFallHorizontal(x, y, 1) && obj == Obj_Falling_Egg) {
				playSound(Sound_Broken_Egg);
				ref(x, y) = Obj_Broken_Egg;
			}
		break;
	case Obj_Wall_North_West:
	case Obj_Filled_Wall_North_West:
		if (!tryFallHorizontal(x, y, -1) && obj == Obj_Falling_Egg) {
			playSound(Sound_Broken_Egg);
			ref(x, y) = Obj_Broken_Egg;
		}
		break;		
	case Obj_Wall_North_East:
	case Obj_Filled_Wall_North_East:
		if (!tryFallHorizontal(x, y, 1) && obj == Obj_Falling_Egg) {
			playSound(Sound_Broken_Egg);
			ref(x, y) = Obj_Broken_Egg;
		}
		break;
	default:
		if (obj == Obj_Falling_Egg) {
			playSound(Sound_Broken_Egg);
			ref(x, y) = Obj_Broken_Egg;
		}
	}
}

// Check if the specified object can fall.
void Map::tryFall(int x, int y)
{
	if (!tryFallDown(x, y))
		tryFallLeftOrRight(x, y);
}

// Check if the blip can go into a direction.
bool Map::blipTryDirection(Monster *m, Orientation dir, int xd, int yd)
{
	switch (xy(m->xpos + xd, m->ypos + yd)) {
	case Obj_Empty:
	case Obj_Ground1:
	case Obj_Ground2:
		m->xpos += xd;
		m->ypos += yd;
		m->orientation = dir;
		return true;
	case Obj_Cage:
		cageBlip(m, m->xpos + xd, m->ypos + yd);
		return true;
	case Obj_Repton:
		playSound(Sound_Die);
		status_die = true;
		break;
	default:;	// Avoid warning
	}

	return false;
}

// Check and move the blip.
void Map::updateBlip(Monster *m)
{
	int mx = m->xpos, my = m->ypos;

	switch (m->orientation) {
	case North:
		if (blipTryDirection(m, West, -1, 0))
			;
		else if (!blipTryDirection(m, North, 0, -1)) {
			m->orientation = East;
			updateBlip(m);
			return;
		}
		break;
	case South:
		if (blipTryDirection(m, East, 1, 0))
			;
		else if (!blipTryDirection(m, South, 0, 1)) {
			m->orientation = West;
			updateBlip(m);
			return;
		}
		break;
	case East:
		if (blipTryDirection(m, North, 0, -1))
			;
		else if (!blipTryDirection(m, East, 1, 0)) {
			m->orientation = South;
			updateBlip(m);
			return;
		}
		break;
	case West:
		if (blipTryDirection(m, South, 0, 1))
			;
		else if (!blipTryDirection(m, West, -1, 0)) {
			m->orientation = North;
			updateBlip(m);
			return;
		}
		break;
	}

	if (m->xpos != mx || m->ypos != my)
		m->sprite = (m->sprite == Obj_Blip) ? Obj_Blip2 : Obj_Blip;
}

// Check if the monster can go into a direction and the distance between
// it and repton.
double Map::monsterTryDirection(Monster *m, int xd, int yd)
{
	switch (xy(m->xpos + xd, m->ypos + yd)) {
	case Obj_Empty:
	case Obj_Ground1:
	case Obj_Ground2:
		return sqrt(pow(m->xpos + xd - xpos, 2) +
			    pow(m->ypos + yd - ypos, 2));
	case Obj_Repton:
		playSound(Sound_Die);
		status_die = true;
	default:;
	}

	return 666;	// A rathen large number
}

// Check the repton position and move the monster according with it.
void Map::updateMonster(Monster *m)
{
	int xd = 0, yd = 0;
	double i, len;

	len = sqrt(pow(m->xpos - xpos, 2) + pow(m->ypos - ypos, 2));

	if ((i = monsterTryDirection(m, 0, -1)) < len) {
		xd = 0; yd = -1; len = i;
	}
	if ((i = monsterTryDirection(m, 0, 1)) < len) {
		xd = 0; yd = 1; len = i;
	}
	if ((i = monsterTryDirection(m, 1, 0)) < len) {
		xd = 1; yd = 0; len = i;
	}
	if (monsterTryDirection(m, -1, 0) < len) {
		xd = -1; yd = 0;
	}
	m->xpos += xd;
	m->ypos += yd;

	m->sprite = (m->sprite == Obj_Monster) ? Obj_Monster2 : Obj_Monster;
}

// Check the broken eggs.
void Map::updateEggs()
{
	for (int y = height - 1; y >= 0; y--)
		for (int x = width - 1; x >= 0; x--)
			switch (xy(x, y)) {
			case Obj_Broken_Egg:
				ref(x, y) = Obj_Broken_Egg1;
				break;
			case Obj_Broken_Egg1:
				ref(x, y) = Obj_Broken_Egg2;
				break;
			case Obj_Broken_Egg2:
				ref(x, y) = Obj_Broken_Egg3;
				break;
			case Obj_Broken_Egg3:
				ref(x, y) = Obj_Empty;
				addMonster(x, y, Obj_Monster);
				break;
			default:;	// Avoid warning
			}
}

// Check the moveable objects (rocks and eggs).
void Map::updateObjects()
{
	// Note the ``height - 2'' value.
	for (int y = height - 2; y >= 0; y--)
		for (int x = width - 1; x >= 0; x--)
			switch(xy(x, y)) {
			case Obj_Rock:
			case Obj_Egg:
			case Obj_Falling_Egg:
			case Obj_Broken_Egg:
			case Obj_Broken_Egg1:
			case Obj_Broken_Egg2:
			case Obj_Broken_Egg3:
				// Check if the object can fall.
				tryFall(x, y);
				break;
			default:;	// Avoid warning
			}
}

// Try to reproduce the plant.
bool Map::tryPlant(int x, int y, int dx, int dy)
{
	if (xy(x+dx, y+dy) == Obj_Empty || xy(x+dx, y+dy) == Obj_Repton) {
		if (xy(x+dx, y+dy) == Obj_Repton) {
			playSound(Sound_Die);
			status_die = true;
		}
		ref(x+dx, y+dy) = Obj_Plant;
		addMonster(x+dx, y+dy, Obj_Plant);
		return true;
	}

	return false;
}

// Check the plants and reproduce.
bool Map::updatePlant(Monster *m)
{
	if (rand() % 100 <= 10)
		if (tryPlant(m->xpos, m->ypos, 1, 0))
			return true;
		else if (tryPlant(m->xpos, m->ypos, -1, 0))
			return true;
		else if (tryPlant(m->xpos, m->ypos, 0, 1))
			return true;
		else if (tryPlant(m->xpos, m->ypos, 0, -1))
			return true;
	return false;
}

// Reproduce the plants
void Map::updatePlants()
{
	for (Monster *m = monsters.first(); m != 0; m = monsters.next())
		if (m->type == Obj_Plant)
			    if (updatePlant(m))
				    break;
}

// Check the monsters status and move them.
void Map::updateMonsters()
{
	for (Monster *m = monsters.first(); m != 0; m = monsters.next())
		switch (m->type) {
		case Obj_Blip:
			updateBlip(m);
			break;
		case Obj_Monster:
			updateMonster(m);
			break;
		default:;	// Avoid warning
		}
}

// Move Repton vertically.
void Map::moveVerticalDirect(int yd)
{
	// Move repton up or down according with the yd value.
	ref(xpos, ypos) = Obj_Empty;
	ref(xpos, ypos + yd) = Obj_Repton;
	ypos += yd;

	// Change the repton sprite orientation.
	switch (sprite) {
	case Obj_Repton:
	case Obj_Repton_Look_Left:
	case Obj_Repton_Look_Right:
	case Obj_Repton_Go_Right1:
	case Obj_Repton_Go_Right2:
	case Obj_Repton_Go_Right3:
	case Obj_Repton_Go_Right4:
	case Obj_Repton_Go_Left1:
	case Obj_Repton_Go_Left2:
	case Obj_Repton_Go_Left3:
	case Obj_Repton_Go_Left4:
	case Obj_Repton_Go_Up2:
		sprite = Obj_Repton_Go_Up1;
		break;
	case Obj_Repton_Go_Up1:
		sprite = Obj_Repton_Go_Up2;
		break;
	default:;	// Avoid warning
	}
}

// Move Repton horizontally.
void Map::moveHorizontalDirect(int xd)
{
	// Move repton to the left or to the right according
	// with the xd value.
	ref(xpos, ypos) = Obj_Empty;
	ref(xpos + xd, ypos) = Obj_Repton;
	xpos += xd;

	// Change the repton sprite orientation.
	if (xd == -1)
		//  Move repton to the left.
		switch (sprite) {
		case Obj_Repton:
		case Obj_Repton_Look_Left:
		case Obj_Repton_Look_Right:
		case Obj_Repton_Go_Up1:
		case Obj_Repton_Go_Up2:
		case Obj_Repton_Go_Right1:
		case Obj_Repton_Go_Right2:
		case Obj_Repton_Go_Right3:
		case Obj_Repton_Go_Right4:
		case Obj_Repton_Go_Left4:
			sprite = Obj_Repton_Go_Left1;
			break;
		case Obj_Repton_Go_Left1:
			sprite = Obj_Repton_Go_Left2;
			break;
		case Obj_Repton_Go_Left2:
			sprite = Obj_Repton_Go_Left3;
			break;
		case Obj_Repton_Go_Left3:
			sprite = Obj_Repton_Go_Left4;
			break;
		default:;	// Avoid warning
		}
	else
		// Move repton to the right.
		switch (sprite) {
		case Obj_Repton:
		case Obj_Repton_Look_Left:
		case Obj_Repton_Look_Right:
		case Obj_Repton_Go_Up1:
		case Obj_Repton_Go_Up2:
		case Obj_Repton_Go_Left1:
		case Obj_Repton_Go_Left2:
		case Obj_Repton_Go_Left3:
		case Obj_Repton_Go_Left4:
		case Obj_Repton_Go_Right4:
			sprite = Obj_Repton_Go_Right1;
			break;
		case Obj_Repton_Go_Right1:
			sprite = Obj_Repton_Go_Right2;
			break;
		case Obj_Repton_Go_Right2:
			sprite = Obj_Repton_Go_Right3;
			break;
		case Obj_Repton_Go_Right3:
			sprite = Obj_Repton_Go_Right4;
			break;
		default:;	// Avoid warning
		}
}

// This function is called when a key is get. It makes the strongbox diamonds.
void Map::gotKey()
{
	if (!have_key) {
		// Automagically replace all the strongboxes with diamonds.
		for (int y = 0; y < height; ++y)
			for (int x = 0; x < width; ++x)
				if (xy(x, y) == Obj_Strongbox)
					ref(x, y) = Obj_Diamond;
		have_key = true;
	}
}

// This function is called when Repton get an object.
void Map::gotObject(SpriteObj obj)
{
	switch (obj) {
	case Obj_Diamond:
		playSound(Sound_Got_Diamond);
		--num_diamonds;
		num_points += 50;
		break;
	case Obj_Time:
		playSound(Sound_Got_Time);
		num_secs += 30;
		num_points += 10;
		break;
	case Obj_Key:
		playSound(Sound_Got_Key);
		gotKey();
		num_points += 30;
		break;
	case Obj_Crown:
		playSound(Sound_Got_Crown);
		num_points += 200;
		break;
	case Obj_Bomb:
		if (num_diamonds == 0)
			status_end_level = true;
		break;
	default:;	// Avoid warning
	}
}

// Move Repton to a new map location when it enters a transport.
void Map::useTransporter(int x, int y)
{
	Transporter *t = findTransporter(x, y);

	// Remove the transporter and Repton from the original location.
	ref(x, y) = Obj_Empty;
	ref(xpos, ypos) = Obj_Empty;

	playSound(Sound_Transport);

	xpos = t->dest_x;
	ypos = t->dest_y;
	switch (xy(xpos, ypos)) {
	case Obj_Diamond:
	case Obj_Time:
	case Obj_Key:
	case Obj_Crown:
		gotObject(xy(xpos, ypos));
		// Move repton to the new location.
		ref(xpos, ypos) = Obj_Repton;
		sprite = Obj_Repton;
		break;
	case Obj_Ground1:
	case Obj_Ground2:
		num_points += 2;
		/* FALLTHROUGH */
	case Obj_Empty:
		if (findMonster(xpos, ypos) != 0) {
			playSound(Sound_Die);
			status_die = true;
		}
		// Move repton to the new location.
		ref(xpos, ypos) = Obj_Repton;
		sprite = Obj_Repton;
		break;
	case Obj_Transport:
		useTransporter(xpos, ypos);
		break;
	default:
		playSound(Sound_Die);
		status_die = true;
	}
}

// Check if Repton can move vertically according with the yd value.
void Map::moveVertical(int yd)
{
	SpriteObj obj = xy(xpos, ypos + yd);

	if (findMonster(xpos, ypos + yd) != 0) {
		playSound(Sound_Die);
		status_die = true;
	}

	switch (obj) {
	case Obj_Diamond:
	case Obj_Time:
	case Obj_Crown:
	case Obj_Key:
		gotObject(obj);
		/* FALLTHROUGH */
	case Obj_Empty:
		moveVerticalDirect(yd);
		break;
	case Obj_Ground1:
	case Obj_Ground2:
		num_points += 2;
		moveVerticalDirect(yd);
		break;
	case Obj_Transport:
		useTransporter(xpos, ypos + yd);
		break;
	case Obj_Skull:
	case Obj_Plant:
		playSound(Sound_Die);
		status_die = true;
		break;
	case Obj_Bomb:
		gotObject(Obj_Bomb);
	default:;	// Avoid warning
	}
}

// Move Repton and a near object horizontally.
void Map::moveHorizontalMoveObj(int xd, SpriteObj obj)
{
	if (xy(xpos + xd*2, ypos) == Obj_Empty) {
		Monster *m;
		if ((m = findMonster(xpos + xd*2, ypos)) != 0 && m->type == Obj_Monster) {
			playSound(Sound_Kill_Monster);
			killMonster(m);
		}
		ref(xpos + xd*2, ypos) = obj;
		moveHorizontalDirect(xd);
	}
}

// Check if Repton can move horizontally according with the xd value.
void Map::moveHorizontal(int xd)
{
	SpriteObj obj = xy(xpos + xd, ypos);

	if (findMonster(xpos + xd, ypos) != 0) {
		playSound(Sound_Die);
		status_die = true;
	}

	switch (obj) {
	case Obj_Rock:
	case Obj_Egg:
	case Obj_Broken_Egg:
	case Obj_Broken_Egg1:
	case Obj_Broken_Egg2:
	case Obj_Broken_Egg3:
		moveHorizontalMoveObj(xd, obj);
		break;
	case Obj_Diamond:
	case Obj_Time:
	case Obj_Crown:
	case Obj_Key:
		gotObject(obj);
		/* FALLTHROUGH */
	case Obj_Empty:
		moveHorizontalDirect(xd);
		break;
	case Obj_Ground1:
	case Obj_Ground2:
		num_points += 2;
		moveHorizontalDirect(xd);
		break;
	case Obj_Transport:
		useTransporter(xpos + xd, ypos);
		break;
	case Obj_Skull:
	case Obj_Plant:
		playSound(Sound_Die);
		status_die = true;
		break;
	case Obj_Bomb:
		gotObject(Obj_Bomb);
	default:;	// Avoid warning
	}
}

// Kill a monster.
void Map::killMonster(Monster *mp)
{
	num_points += 500;
	for (Monster *m = monsters.first(); m != 0; m = monsters.next())
		if (m == mp)
			monsters.remove();
}

// Kill a blip, make it a diamond, and produce a sound.
void Map::cageBlip(Monster *m, int x, int y)
{
	playSound(Sound_Cage_Blip);
	ref(x, y) = Obj_Diamond;
	killMonster(m);
}
